/*
    alarmwatch - lm_sensors alarm watching daemon
    Copyright (C) 2001  Matthew Mueller <donut@azstarnet.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

//todo:
//setuid(nobody) after all files opened and stuff? probably not really needed since no untrusted input is taken, but I guess it could be handy incase configure found the wrong SENSORS somehow? (also would need to regain original uid to unlink pidfile...)

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

//get asprintf()
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include <errno.h>

#include <signal.h>
#include <syslog.h>

#define PROGRAM "alarmwatch"
#define VERSION "1.0"

#define SENSORS_DIR "/proc/sys/dev/sensors"
//#define SENSORS "/usr/local/bin/sensors"

FILE * beepdev=NULL;
char * pidfile=NULL;

typedef struct Watch {
	char *file;
	char *name;
	int ignore;
} Watch;



typedef void (*syslog_funcp)(int, const char *, ...);
void fakesyslog(int priority, const char *format, ...){
	va_list ap;
	char buf[20];

	struct tm *time_now;
	time_t t;

	time(&t);
	time_now = localtime(&t);
	if (strftime(buf,20,"%Y-%m-%d %H:%M:%S",time_now)>0)
		printf("%s ",buf);

	va_start(ap,format);
	vprintf(format,ap);
	va_end(ap);
	printf("\n");
	fflush(stdout);
}

syslog_funcp syslogfn=fakesyslog;

//#define FCHK(a,n,f) if ((a)<0) {fprintf(stderr,"%s %s: %s\n",#n,f,strerror(errno));exit(1);}
#define FCHK(a,n,f) fchk((a)<0,n,f)
void fchk(int a,char *n, char *f) {
	if (a) {
		syslogfn(LOG_ERR,"%s %s: %s",n,f,strerror(errno));
		exit(1);
	}
}
int atoichk(char *a,char *n, int min) {
	char *ep;
	int i=strtol(a,&ep,0);
	if (*a=='\0' || *ep!='\0' || i<min) {
		syslogfn(LOG_ERR,"invalid %s: %s",n,a);
		exit(1);
	}
	return i;
}


void dobeep(void){
	if (beepdev){
		if (fputc('\a',beepdev)!='\a'){
			syslogfn(LOG_ERR,"%s %s: %s","fputc","beepdev",strerror(errno));
			beepdev=NULL;
		}else
			fflush(beepdev);
	}
}



void print_version(void){
	printf("%s\n",VERSION);
}
void print_help(void){
	printf("%s v%s Copyright (c) 2001 Matt Mueller <donut@azstarnet.com>\n", PROGRAM, VERSION);
	printf("see `man %s' for usage\n",PROGRAM);
}

void cleanup(int sig){
	syslogfn(LOG_INFO,"%s v%s exiting on sig %i",PROGRAM,VERSION,sig);
	if (pidfile)
		FCHK(unlink(pidfile),"unlink",pidfile);
	exit(0);
}

#define BUFSIZE 81
int main(int argc,char **argv){
	char buf[BUFSIZE];
	int v,vi,i;
	int numwatch=0;
	Watch *watch=NULL;
	Watch *w;
	int fh;
	int doalarm=0;
	FILE *fp;
	char *strp;
	time_t nextbeep=0, nextlog=0, now;

	int usesyslog=1;
	int debug=0;
	int daemonic=1;
	int beepinterval=30;
	int loginterval=60;
	int checkinterval=5;

	struct option long_opts[] =  {
		{ "help", no_argument, NULL, 'h' },
		{ "version", no_argument, NULL, 'v'},
		{ "beepdev", required_argument, NULL, 'b' },
		{ "beepinterval", required_argument, NULL, 'B' },
		{ "loginterval", required_argument, NULL, 'L' },
		{ "checkinterval", required_argument, NULL, 'C' },
		{ "pidfile", required_argument, NULL, 'p' },
		{ "nosyslog", no_argument, NULL, 'S' },
		{ "debug", no_argument, NULL, 'd' },
		{ "nodaemon", no_argument, NULL, 'D' },
		{ 0,0,0,0 }
	};

	while (1) {
		v = getopt_long(argc,argv,"hvb:B:L:C:p:dSD",long_opts,NULL);
		if (v == EOF)
			break;
		switch(v) {
			case ':':
			case '?':
			case 'h':
				print_help();
				exit(0);
			case 'v':
				print_version();
				exit(0);
			case 'b':
				fchk((beepdev=fopen(optarg,"a"))==NULL,"fopen",optarg);
				break;
			case 'B':
				beepinterval=atoichk(optarg,"beepinterval",0);
				break;
			case 'L':
				loginterval=atoichk(optarg,"loginterval",0);
				break;
			case 'C':
				checkinterval=atoichk(optarg,"checkinterval",1);
				break;
			case 'p':
				pidfile=optarg;
				break;
			case 'S':
				usesyslog = 0;
				break;
			case 'd':
				debug = 1;
				break;
			case 'D':
				daemonic = 0;
				break;
			default:
				fprintf(stderr,"getopt error (v=%i)!\n",v);
				exit(1);
		}
	}
	if (!beepinterval && !loginterval){
		fprintf(stderr,"if you aren't going to beep, and you aren't going to log, why should I even bother?\n");
		exit(1);
	}

	vi=argc-optind;
	if (vi<=0){
		print_help();
		fprintf(stderr,"no chips specified\n");
		exit(1);
	}
	if (vi%2){
		fprintf(stderr,"uneven number of args specified\n");
		exit(1);
	}
	numwatch=vi/2;
	watch=(Watch*)malloc(sizeof(Watch)*numwatch);//we don't care about this freeing memory, since we need it till the program exits anyway.
	for (i=0;i<numwatch;i++){
		v=optind+i*2;
		asprintf(&watch[i].file,"%s/alarms",argv[v]);//allocates memory automatically, don't care about freeing this either.
		watch[i].name=argv[v];
		watch[i].ignore=atoichk(argv[v+1],"ignorespec",0);
		if (debug)
			printf("%i: %s %s %i\n",i,watch[i].file,watch[i].name,watch[i].ignore);
	}

	FCHK(chdir(SENSORS_DIR),"chdir",SENSORS_DIR);

	if (beepinterval && !beepdev){
		beepdev=fopen("/dev/console","a");
		if (!beepdev){
			if (debug) syslogfn(LOG_DEBUG,"beepdev not specified and couldn't open /dev/console, using stderr.");
			beepdev=stderr;
		}
	}

	if (daemonic){
		/*FCHK(v=fork(),"fork","");
		if (v!=0) {
			exit(0);
		}*/
		int noclose=1;
		if ((beepdev != stderr) && usesyslog)
			noclose=0;
		if (debug)
			syslogfn(LOG_DEBUG,"noclose: %i",noclose);
		FCHK(daemon(1,noclose),"daemon",noclose?"1":"0");
	}

	if (usesyslog) {
		openlog(PROGRAM, LOG_CONS, LOG_DAEMON);
		syslogfn=syslog;
	}

	if (pidfile) {
		FCHK(fh=open(pidfile,O_WRONLY|O_CREAT|O_EXCL,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH),"open",pidfile); //use open instead of fopen so we can do O_EXCL.
		i=sprintf(buf,"%i\n",getpid());
		FCHK(write(fh,buf,i),"write",pidfile);
		FCHK(close(fh),"close",pidfile);
	}

	signal(SIGINT,cleanup);
	signal(SIGQUIT,cleanup);
	signal(SIGTERM,cleanup);
	syslogfn(LOG_INFO,"%s v%s started",PROGRAM,VERSION);

	sleep(2);//delay a bit before first read, so that sensors -s in init file will have time to take effect and any erroneous alarms will be cleared.
	while (1){
		now=time(NULL);

		for (i=0;i<numwatch;i++){
			w=watch+i;
			FCHK(fh=open(w->file,O_RDONLY),"open",w->file);
			FCHK(v=read(fh,buf,BUFSIZE-1),"read",w->file);
			FCHK(close(fh),"close",w->file);
			buf[v]=0;
			v=atoi(buf);
			vi = v & ~w->ignore;
			if (debug){
				syslogfn(LOG_DEBUG,"%s: v=%i vi=%i",w->file,v,vi);
			}
			if (vi) {
				if (loginterval && now>=nextlog){
					doalarm=2;
					syslogfn(LOG_CRIT,"%s: %i(0x%x) (%i & ~%i)!",w->file,vi,vi,v,w->ignore);
					//sprintf(buf,SENSORS" %s | "GREP" ALARM",w->name);
					//system(buf);
					sprintf(buf,SENSORS" %s",w->name);
					fchk((fp=popen(buf,"r"))==NULL,"popen",buf);
					while (fgets(buf,BUFSIZE,fp)){
						if (strstr(buf,"ALARM")){
							if ((strp=strrchr(buf,'\n')))
								*strp=0;
							syslogfn(LOG_CRIT,"%s",buf);
						}
					}
					pclose(fp);
				}else
					doalarm=1;
			}
		}
		if (doalarm){
			if (doalarm==2)
				nextlog=now+loginterval;
			if (beepinterval && now>=nextbeep){
				nextbeep=now+beepinterval;
				if (debug){
					syslogfn(LOG_DEBUG,"beep!");
				}else{
					dobeep();
					usleep(75000);
					dobeep();
					sleep(1);
					dobeep();
					usleep(75000);
					dobeep();
				}
			}
			now=time(NULL);
			if (loginterval && (!beepinterval || nextbeep>nextlog))
				sleep(nextlog-now);
			else
				sleep(nextbeep-now);
			doalarm=0;
		}else
			sleep(checkinterval);
	}
	return 0;
}
